#ifndef PPM_EXTENTS_H
#define PPM_EXTENTS_H
/**
 * @file ppm_extents.h
 * @brief declarations for functions on extents
 *
 * @copyright Copyright  (C)  2011,2012  Thomas Jahns <jahns@dkrz.de>
 *
 * @version 1.0
 * @author Thomas Jahns <jahns@dkrz.de>
 */
/*
 * Maintainer: Thomas Jahns <jahns@dkrz.de>
 * URL: https://www.dkrz.de/redmine/projects/show/scales-ppm
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <inttypes.h>
#include <stdlib.h>

/**
 * integer range description
 */
struct PPM_extent
{
  int32_t first, /**< the range is anchored at \a first */
    size;  /**< the range has size \a size, i.e. the range equals
            * \li [first, first + size - 1] for size > 0
            * \li [first + size + 1, first] for size < 0
            * \li the empty set if size = 0
            */
};

static inline int32_t
PPM_extent_size(struct PPM_extent extent)
{
  return extent.size;
}

static inline int32_t
PPM_extents_size(size_t nextents, const struct PPM_extent extent[])
{
  int32_t size = 1;
  if (!nextents)
    return 0;
  for (int i = 0; i < nextents; ++i)
    size *= PPM_extent_size(extent[i]);
  return size;
}

static inline int32_t
PPM_extent_start(struct PPM_extent extent)
{
  return extent.first;
}

static inline int32_t
PPM_extent_end(struct PPM_extent extent)
{
  return extent.first + extent.size  - ((extent.size >= 0) * 2 - 1);
}

static inline int
PPM_extent_eq(struct PPM_extent a, struct PPM_extent b)
{
  return (a.first == b.first) && (a.size == b.size);
}

static inline int
PPM_extents_eq(size_t nextents, struct PPM_extent a[nextents],
               struct PPM_extent b[nextents])
{
  size_t i;
  for (i = 0; i < nextents && PPM_extent_eq(a[i], b[i]); ++i)
    ;
  return i == nextents;
}

/**
 * inclusive integer range description
 *
 * describes range [first,last]
 */
struct PPM_iinterval
{
  int32_t first,   /**< the range is anchored at \a first */
    last;      /**< the range contains \a last */
};

static inline int32_t
PPM_iinterval_size(struct PPM_iinterval interval)
{
  int32_t rng_size = interval.last - interval.first;
  return rng_size + ((rng_size >= 0) * 2 - 1);
}

static inline int32_t
PPM_iintervals_size(size_t nintervals, const struct PPM_iinterval intervals[])
{
  int32_t size = 1;
  if (!nintervals)
    return 0;
  for (int i = 0; i < nintervals; ++i)
    size *= PPM_iinterval_size(intervals[i]);
  return size;
}

static inline int32_t
PPM_iinterval_start(struct PPM_iinterval iinterval)
{
  return iinterval.first;
}

static inline int32_t
PPM_iinterval_end(struct PPM_iinterval iinterval)
{
  return iinterval.last;
}

static inline struct PPM_iinterval
PPM_extent2iinterval(const struct PPM_extent src)
{
  struct PPM_iinterval dst =
    { .first = src.first, .last = PPM_extent_end(src) };
  return dst;
}

void
PPM_extents2iintervals(int ndims, struct PPM_iinterval dst[ndims],
                       const struct PPM_extent src[ndims]);

/**
 * inclusive float range description
 *
 * describes range [first,last]
 */
struct PPM_iinterval_sp
{
  float first,  /**< the range is anchored at \a first */
    last;       /**< the range contains \a last */
};

/**
 * inclusive double range description
 *
 * describes range [first,last]
 */
struct PPM_iinterval_dp
{
  double first, /**< the range is anchored at \a first */
    last;       /**< the range contains \a last */
};

/**
 * Write string representation of \a extent to \a buf.
 * @param buf must point to array at least 26 char long.
 * @param ext range to print
 */
void
PPM_sprint_extent(char buf[], struct PPM_extent *ext);

/*
 * Local Variables:
 * license-project-url: "https://www.dkrz.de/redmine/projects/show/scales-ppm"
 * license-markup: "doxygen"
 * license-default: "bsd"
 * End:
 */
#endif
