! @file ppm_math_extensions.f90
! @brief utility routines for floating-point math
!
! Copyright  (C)  2012  Thomas Jahns <jahns@dkrz.de>
!
! @version 1.0
! Keywords:
! @author Thomas Jahns <jahns@dkrz.de>
! Maintainer: Thomas Jahns <jahns@dkrz.de>
! URL: https://www.dkrz.de/redmine/projects/show/scales-ppm
!
! Redistribution and use in source and binary forms, with or without
! modification, are  permitted provided that the following conditions are
! met:
!
! Redistributions of source code must retain the above copyright notice,
! this list of conditions and the following disclaimer.
!
! Redistributions in binary form must reproduce the above copyright
! notice, this list of conditions and the following disclaimer in the
! documentation and/or other materials provided with the distribution.
!
! Neither the name of the DKRZ GmbH nor the names of its contributors
! may be used to endorse or promote products derived from this software
! without specific prior written permission.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
! IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
! TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
! PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
! OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
! EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
! PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
! PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
! LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
! NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
! SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
!
#ifdef __xlC__
@PROCESS STRICT(precision)
#endif
MODULE ppm_math_extensions
#ifdef USE_MPI
  USE ppm_base, ONLY: abort_ppm, ppm_default_comm
  USE ppm_std_type_kinds_mp, ONLY: mp_dp
#endif
  USE ppm_std_type_kinds, ONLY: dp, sp, i4
  USE ppm_math_extensions_internal, ONLY: fpu_save_cw => ppm_fpu_save_cw, &
       fpu_restore_cw => ppm_fpu_restore_cw, &
       fpu_set_precision, fpu_precision_sp, fpu_precision_dp, fpu_precision_ep
  IMPLICIT NONE
  PRIVATE
#ifdef USE_MPI
  INCLUDE 'mpif.h'
#endif
  REAL(dp), PARAMETER, PUBLIC :: &
       m_pi_dp=3.14159265358979323846_dp
  REAL(sp), PARAMETER, PUBLIC :: &
       m_pi_sp=3.14159265358979323846_sp

  !> data edit descriptors for real kinds sp and dp
  CHARACTER(len=10) :: de_g_sp, de_g_dp

  !   The ddp functions compute complex results that store a Kahan
  !   correction term in the imaginary part
  INTERFACE ddp_add
    MODULE PROCEDURE ddp_add_ddp_ddp
    MODULE PROCEDURE ddp_add_dp_dp
  END INTERFACE ddp_add
  INTERFACE ddp_sum
    MODULE PROCEDURE ddp_sum_dp_1d
    MODULE PROCEDURE ddp_sum_dp_2d
    MODULE PROCEDURE ddp_sum_dp_3d
  END INTERFACE ddp_sum
#ifdef USE_MPI
  INTEGER :: ddpdd_op
  INTERFACE ddp_sum_mp
    MODULE PROCEDURE ddp_sum_mp_dp_1d
    MODULE PROCEDURE ddp_sum_mp_dp_2d
    MODULE PROCEDURE ddp_sum_mp_dp_3d
  END INTERFACE ddp_sum_mp
#endif
  PUBLIC :: ddp_add, ddp_sum, ddp_abs
  PUBLIC :: fpu_save_cw, fpu_restore_cw, fpu_set_precision, &
       fpu_precision_sp, fpu_precision_dp, &
       fpu_precision_ep
  PUBLIC :: initialize_math_extensions
  PUBLIC :: de_g_sp, de_g_dp
CONTAINS
  !   Modification of original codes written by David H. Bailey
  !>  This function computes c = a(i)+b(i) but gives as result
  !!  a complex that stores a Kahan correction term in the imaginary part
  ELEMENTAL FUNCTION ddp_add_ddp_ddp(a, b) RESULT(c)
    COMPLEX(dp) :: c
    COMPLEX(dp), INTENT(in) :: a, b
    REAL(dp) :: t1, t2, e
#ifdef NEED_PRECISION_CONTROL
    INTEGER(i4) :: fpu_cw
    CALL fpu_set_precision(fpu_precision_dp, fpu_cw)
#endif
    t1 = REAL(a, dp) + REAL(b, dp)
    e = t1 - REAL(a, dp)
    t2 = ((REAL(b, dp) - e) + (REAL(a, dp) - (t1 - e))) &
         + AIMAG(a) + AIMAG(b)
    c = CMPLX(t1 + t2, t2 - ((t1 + t2) - t1), dp)
#ifdef NEED_PRECISION_CONTROL
    CALL fpu_restore_cw(fpu_cw)
#endif
  END FUNCTION ddp_add_ddp_ddp

  ELEMENTAL FUNCTION ddp_add_dp_dp(a, b) RESULT(c)
    COMPLEX(dp) :: c
    REAL(dp), INTENT(in) :: a, b
    REAL(dp) :: t1, t2, e
#ifdef NEED_PRECISION_CONTROL
    INTEGER(i4) :: fpu_cw
    CALL fpu_set_precision(fpu_precision_dp, fpu_cw)
#endif
    t1 = a + b
    e = t1 - a
    t2 = (b - e) + (a - (t1 - e))
    c = CMPLX(t1 + t2, t2 - ((t1 + t2) - t1), dp)
#ifdef NEED_PRECISION_CONTROL
    CALL fpu_restore_cw(fpu_cw)
#endif
  END FUNCTION ddp_add_dp_dp

  ELEMENTAL FUNCTION ddp_abs(a) RESULT(c)
    COMPLEX(dp) :: c
    COMPLEX(dp), INTENT(in) :: a
    c = MERGE(a, -a, REAL(a, dp) >= 0.0_dp)
  END FUNCTION ddp_abs

  !> compute double-double-precision corrected sum of 1d-array, should
  !! give better results than sum(a) for arrays a where cancellation
  !! occurs.
  !! @param a array a(1)..a(n) to sum up
  !! @return \f$\sum^n_{i=1} a_i\f$
  PURE FUNCTION ddp_sum_dp_1d(a) RESULT(c)
    COMPLEX(dp) :: c
    REAL(dp), INTENT(in) :: a(:)
    REAL(dp) :: t1, t2, e
    INTEGER :: len, i
#ifdef NEED_PRECISION_CONTROL
    INTEGER(i4) :: fpu_cw
    CALL fpu_set_precision(fpu_precision_dp, fpu_cw)
#endif
    c = (0.0_dp, 0.0_dp)
    len = SIZE(a)
    DO i = 1, len
      t1 = a(i) + REAL(c, dp)
      e = t1 - a(i)
      t2 = ((REAL(c, dp) - e) + (a(i) - (t1 - e))) + AIMAG(c)
      c = CMPLX(t1 + t2, t2 - ((t1 + t2) - t1), dp)
    END DO
#ifdef NEED_PRECISION_CONTROL
    CALL fpu_restore_cw(fpu_cw)
#endif
  END FUNCTION ddp_sum_dp_1d

  !> compute double-double-precision corrected sum of 2d-array
  !> @see ddp_sum_dp_1d
  !> @param a array a(1,1)..a(m,n) to sum up
  !> @return \f$\sum^m_{i=1}\sum^n_{j=1} a_{i,j}\f$
  PURE FUNCTION ddp_sum_dp_2d(a) RESULT(c)
    COMPLEX(dp) :: c
    REAL(dp), INTENT(in) :: a(:, :)
    REAL(dp) :: a1(1)
    c = ddp_sum_dp_1d(TRANSFER(a, a1))
  END FUNCTION ddp_sum_dp_2d

  !> compute double-double-precision corrected sum of 3d-array
  !> @see ddp_sum_dp_1d
  !> @param a array a(1,1,1)..a(m,n,o) to sum up
  !> @return \f$\sum^m_{i=1}\sum^n_{j=1}\sum^o_{l=1} a_{i,j,l}\f$
  PURE FUNCTION ddp_sum_dp_3d(a) RESULT(c)
    COMPLEX(dp) :: c
    REAL(dp), INTENT(in) :: a(:, :, :)
    REAL(dp) :: a1(1)
    c = ddp_sum_dp_1d(TRANSFER(a, a1))
  END FUNCTION ddp_sum_dp_3d

#ifdef USE_MPI
  SUBROUTINE ddpdd(a, b, len, itype)
    INTEGER, INTENT(in) :: len, itype
    COMPLEX(dp), INTENT(in) :: a(len)
    COMPLEX(dp), INTENT(inout) :: b(len)
    SELECT CASE (itype)
    CASE(mp_dp)
      b = ddp_add(a, b)
    CASE default
      CALL abort_ppm("invalid MPI data type in double-double summation", &
           __FILE__, &
           __LINE__)
    END SELECT
  END SUBROUTINE ddpdd

  !> compute double-double-precision corrected sum of distributed 1d-array
  !!
  !! This is a collective MPI-Call.
  !! @param a array a(1)..a(n) to sum up
  !! @return \f$\sum^n_{i=1} a_i\f$
  FUNCTION ddp_sum_mp_dp_1d(a, comm) RESULT(c)
    COMPLEX(dp) :: c
    REAL(dp), INTENT(in) :: a(:)
    INTEGER, INTENT(in) :: comm
    INTEGER :: ierror
    c = ddp_sum(a)
    CALL mpi_allreduce(mpi_in_place, c, 1, mp_dp, ddpdd_op, comm, ierror)
    CALL handle_mpi_error(ierror, comm, &
         __FILE__, &
         __LINE__)
  END FUNCTION ddp_sum_mp_dp_1d

  !> compute double-double-precision corrected sum of distributed 2d-array
  !> @see ddp_sum_mp_dp_1d
  !> @param a array a(1,1)..a(m,n) to sum up
  !> @return \f$\sum^m_{i=1}\sum^n_{j=1} a_{i,j}\f$
  FUNCTION ddp_sum_mp_dp_2d(a, comm) RESULT(c)
    COMPLEX(dp) :: c
    REAL(dp), INTENT(in) :: a(:, :)
    INTEGER, INTENT(in) :: comm
    REAL(dp) :: a1(1)
    c = ddp_sum_mp_dp_1d(TRANSFER(a, a1), comm)
  END FUNCTION ddp_sum_mp_dp_2d

  !> compute double-double-precision corrected sum of distributed 3d-array
  !> @see ddp_sum_mp_dp_1d
  !> @param a array a(1,1,1)..a(m,n,o) to sum up
  !> @return \f$\sum^m_{i=1}\sum^n_{j=1}\sum^o_{l=1} a_{i,j,l}\f$
  FUNCTION ddp_sum_mp_dp_3d(a, comm) RESULT(c)
    COMPLEX(dp) :: c
    REAL(dp), INTENT(in) :: a(:, :, :)
    INTEGER, INTENT(in) :: comm
    REAL(dp) :: a1(1)
    c = ddp_sum_mp_dp_1d(TRANSFER(a, a1), comm)
  END FUNCTION ddp_sum_mp_dp_3d
#endif

  SUBROUTINE initialize_math_extensions
    INTEGER :: field_w, field_d, field_e, min_e, max_e
    REAL(sp) :: f
    REAL(dp) :: d
#ifdef USE_MPI
    INTEGER :: ierror
#endif
    ! width of fraction
    field_d = MERGE(DIGITS(f), &
         CEILING(1.0 + REAL(DIGITS(f)) * LOG10(REAL(RADIX(f)))), &
         RADIX(f) == 10)
    max_e = MAXEXPONENT(f)
    min_e = MINEXPONENT(f)
    ! width of exponent field
    field_e = MAX(&
         CEILING(LOG10(REAL(max_e))), &
         1 + CEILING(LOG10(ABS(REAL(min_e)))))
    ! extra space: decimal separator and literal 'e'
    field_w = 1 + field_d + 1 + field_e
    WRITE(de_g_sp, '(2(a,i0))') 'g', field_w, '.', field_d

    ! width of fraction
    field_d = MERGE(DIGITS(d), &
         CEILING(1.0 + REAL(DIGITS(d)) * LOG10(REAL(RADIX(d)))), &
         RADIX(d) == 10)
    max_e = MAXEXPONENT(d)
    min_e = MINEXPONENT(d)
    ! width of exponent field
    field_e = MAX(&
         CEILING(LOG10(REAL(max_e))), &
         1 + CEILING(LOG10(ABS(REAL(min_e)))))
    ! extra space: decimal separator and literal 'e'
    field_w = 1 + field_d + 1 + field_e
    WRITE(de_g_dp, '(2(a,i0))') 'g', field_w, '.', field_d

#ifdef USE_MPI
    CALL mpi_op_create(ddpdd, .TRUE., ddpdd_op, ierror)
    CALL handle_mpi_error(ierror, ppm_default_comm, &
         __FILE__, &
         __LINE__)
#endif

  END SUBROUTINE initialize_math_extensions

#ifdef USE_MPI
  SUBROUTINE handle_mpi_error(ierror, comm, source, line)
    INTEGER, INTENT(in) :: ierror, comm, line
    CHARACTER, INTENT(in) :: source

    INTEGER, PARAMETER :: max_msg_len=1024
    CHARACTER(len=max_msg_len) :: msg

    IF (ierror /= MPI_SUCCESS) THEN
      CALL mpi_error_string(ierror, msg, max_msg_len, ierror)
      CALL abort_ppm(msg, source, line, comm)
    END IF
  END SUBROUTINE handle_mpi_error
#endif
END MODULE ppm_math_extensions
!
! Local Variables:
! license-markup: "doxygen"
! license-project-url: "https://www.dkrz.de/redmine/projects/show/scales-ppm"
! license-default: "bsd"
! End:
!
